##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
__all__ = ["CanonicalGate"]

import numpy as np
from qiskit.extensions import UnitaryGate


class CanonicalGate(UnitaryGate):
    """
    Class for representing a canonical gate

    ::

        CAN(alpha, beta, gamma) = [[(exp(i*(alpha+beta-gamma)/2)+exp(i*(alpha-beta+gamma)/2))/2, 0,                                                                    0,                                                                    (exp(i*(alpha-beta+gamma)/2)-exp(i*(alpha+beta-gamma)/2))/2],
                                   [0,                                                                 (exp(i*(alpha+beta+gamma)/(-2))+exp(i*(beta+gamma-alpha)/2))/2, (exp(i*(alpha+beta+gamma)/(-2))-exp(i*(beta+gamma-alpha)/2))/2, 0                                                                ],
                                   [0,                                                                 (exp(i*(alpha+beta+gamma)/(-2))-exp(i*(beta+gamma-alpha)/2))/2, (exp(i*(alpha+beta+gamma)/(-2))+exp(i*(beta+gamma-alpha)/2))/2, 0                                                                ],
                                   [(exp(i*(alpha-beta+gamma)/2)-exp(i*(alpha+beta-gamma)/2))/2, 0,                                                                    0,                                                                    (exp(i*(alpha+beta-gamma)/2)+exp(i*(alpha-beta+gamma)/2))/2]]

    """  # noqa: E501

    def __init__(self, alpha: float, beta: float, gamma: float):
        """
        Args:
            alpha: X-axis phase angle
            beta: Y-axis phase angle
            gamma: Z-axis phase angle
        """
        # fmt: off
        super().__init__(
            np.array([[(np.exp(1j * (alpha + beta - gamma) / 2) + np.exp(1j * (alpha - beta + gamma) / 2)) / 2, 0,                                                                                          0,                                                                                          (np.exp(1j * (alpha - beta + gamma) / 2) - np.exp(1j * (alpha + beta - gamma) / 2)) / 2],    # noqa: E501, E241, E202
                      [0,                                                                                       (np.exp(1j * (alpha + beta + gamma) / (-2)) + np.exp(1j * (beta + gamma - alpha) / 2)) / 2, (np.exp(1j * (alpha + beta + gamma) / (-2)) - np.exp(1j * (beta + gamma - alpha) / 2)) / 2, 0                                                                                      ],    # noqa: E501, E241, E202
                      [0,                                                                                       (np.exp(1j * (alpha + beta + gamma) / (-2)) - np.exp(1j * (beta + gamma - alpha) / 2)) / 2, (np.exp(1j * (alpha + beta + gamma) / (-2)) + np.exp(1j * (beta + gamma - alpha) / 2)) / 2, 0                                                                                      ],    # noqa: E501, E241, E202
                      [(np.exp(1j * (alpha - beta + gamma) / 2) - np.exp(1j * (alpha + beta - gamma) / 2)) / 2, 0,                                                                                          0,                                                                                          (np.exp(1j * (alpha + beta - gamma) / 2) + np.exp(1j * (alpha - beta + gamma) / 2)) / 2]]),  # noqa: E501, E241, E202
            "can"
        )
        # fmt: on



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Optional, Any, Union, List
from uuid import uuid4

from pyquil import get_qc
from pyquil.api import QuantumComputer, EngagementManager
from qcs_api_client.client import QCSClientConfiguration
from qiskit import QuantumCircuit, ClassicalRegister
from qiskit.circuit import Measure
from qiskit.providers import BackendV1, Options, Provider
from qiskit.providers.models import QasmBackendConfiguration

from ._qcs_job import RigettiQCSJob


def _prepare_readouts(circuit: QuantumCircuit) -> None:
    """
    Errors if measuring into more than one readout. If only measuring one, ensures its name is 'ro'. Mutates the input
    circuit.
    """

    # cache register locations for each bit in circuit
    bit_info = {bit: {"reg": reg, "idx": i} for reg in circuit.cregs for i, bit in enumerate(reg)}

    # NOTE: each measure is a tuple of the form (measure, qubits, classical bits)
    measures = [d for d in circuit.data if isinstance(d[0], Measure)]
    readout_names = list({bit_info[clbit]["reg"].name for m in measures for clbit in m[2]})
    num_readouts = len(readout_names)

    if num_readouts == 0:
        raise RuntimeError("Circuit has no measurements")

    if num_readouts > 1:
        readout_names.sort()
        raise RuntimeError(
            f"Multiple readout registers are unsupported on QCSBackend; found {', '.join(readout_names)}"
        )

    orig_readout_name = readout_names[0]
    if orig_readout_name == "ro":
        return

    for i, reg in enumerate(circuit.cregs):
        if reg.name != orig_readout_name:
            continue

        # rename register to "ro"
        ro_reg = ClassicalRegister(size=reg.size, name="ro")
        circuit.cregs[i] = ro_reg

        # fix classical bit references in circuit
        for i, clbit in enumerate(circuit.clbits):
            orig_reg = bit_info[clbit]["reg"]
            if orig_reg.name == orig_readout_name:
                idx = bit_info[clbit]["idx"]
                circuit.clbits[i] = ro_reg[idx]

        # fix classical bit references in measures
        for m in measures:
            for i, clbit in enumerate(m[2]):
                orig_reg = bit_info[clbit]["reg"]
                if orig_reg.name == orig_readout_name:
                    idx = bit_info[clbit]["idx"]
                    m[2][i] = ro_reg[idx]
        break


def _prepare_circuit(circuit: QuantumCircuit) -> QuantumCircuit:
    """
    Returns a prepared copy of the circuit for execution on the QCS Backend.
    """
    circuit = circuit.copy()
    _prepare_readouts(circuit)
    return circuit


class RigettiQCSBackend(BackendV1):
    """
    Class for representing a Rigetti backend, which may target a real QPU or a simulator.
    """

    def __init__(
        self,
        *,
        compiler_timeout: float,
        execution_timeout: float,
        client_configuration: QCSClientConfiguration,
        engagement_manager: EngagementManager,
        backend_configuration: QasmBackendConfiguration,
        provider: Optional[Provider],
        **fields: Any,
    ) -> None:
        """
        Args:
            execution_timeout: Time limit for execution requests, in seconds.
            compiler_timeout: Time limit for compiler requests, in seconds.
            client_configuration: QCS client configuration.
            engagement_manager: QPU engagement manager.
            backend_configuration: Backend configuration.
            provider: Parent provider.
            fields: Keyword arguments for the values to use to override the default options.
        """
        super().__init__(backend_configuration, provider, **fields)
        self._compiler_timeout = compiler_timeout
        self._execution_timeout = execution_timeout
        self._client_configuration = client_configuration
        self._engagement_manager = engagement_manager
        self._qc: Optional[QuantumComputer] = None

    @classmethod
    def _default_options(cls) -> Options:
        return Options(shots=None)

    def run(
        self,
        run_input: Union[QuantumCircuit, List[QuantumCircuit]],
        **options: Any,
    ) -> RigettiQCSJob:
        """
        Run the quantum circuit(s) using this backend.

        Args:
            run_input: Either a single :class:`QuantumCircuit` to run or a list of them to run in parallel.
            **options: Execution options to forward to :class:`RigettiQCSJob`.

        Returns:
            RigettiQCSJob: The job that has been started. Wait for it by calling :func:`RigettiQCSJob.result`
        """
        if not isinstance(run_input, list):
            run_input = [run_input]

        bindings = options.get("parameter_binds") or []
        if len(bindings) > 0:
            run_input = [circuit.bind_parameters(binding) for circuit in run_input for binding in bindings]

        run_input = [_prepare_circuit(circuit) for circuit in run_input]

        if self._qc is None:
            self._qc = get_qc(
                self.configuration().backend_name,
                compiler_timeout=self._compiler_timeout,
                execution_timeout=self._execution_timeout,
                client_configuration=self._client_configuration,
                engagement_manager=self._engagement_manager,
            )

        return RigettiQCSJob(
            job_id=str(uuid4()),
            circuits=run_input,
            options=options,
            qc=self._qc,
            backend=self,
            configuration=self.configuration(),
        )



from typing import Callable

PreCompilationHook = Callable[[str], str]
"""Represents a function that can transform a QASM program string just before compilation."""


def set_rewiring(rewiring: str) -> PreCompilationHook:
    """
    Create a hook which will apply rewiring before compilation.

    See: https://pyquil-docs.rigetti.com/en/stable/compiler.html#initial-rewiring for more information.

    Args:
        rewiring: Rewiring directive to apply.

    Returns:
        PreCompilationHook: A hook to apply rewiring.

    Examples:
        Applying rewiring to a program::

            >>> from qiskit import execute
            >>> from qiskit_rigetti import RigettiQCSProvider, QuilCircuit
            >>> from qiskit_rigetti.hooks.pre_compilation import set_rewiring

            >>> p = RigettiQCSProvider()
            >>> backend = p.get_simulator(num_qubits=2, noisy=True)
            >>> circuit = QuilCircuit(2, 2)
            >>> _ = circuit.measure([0, 1], [0, 1])
            >>> job = execute(circuit, backend, shots=10, before_compile=[set_rewiring("NAIVE")])
    """

    def fn(qasm: str) -> str:
        return qasm.replace("OPENQASM 2.0;", f'OPENQASM 2.0;\n#pragma INITIAL_REWIRING "{rewiring}";')

    return fn



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
__all__ = [
    "CPhase00Gate",
    "CPhase01Gate",
    "CPhase10Gate",
]

from pyquil.simulation.matrices import CPHASE00, CPHASE01, CPHASE10
from qiskit.extensions import UnitaryGate


class CPhase00Gate(UnitaryGate):
    """
    Class for representing a CPhase00 gate, a variant of CPhase that affects state ``|00>``

    ::

        CPHASE00(theta) = [[exp(i*theta), 0, 0, 0],
                           [0,            1, 0, 0],
                           [0,            0, 1, 0],
                           [0,            0, 0, 1]]

    """

    def __init__(self, theta: float):
        """
        Args:
            theta: Phase angle
        """
        super().__init__(CPHASE00(theta), "cphase00")


class CPhase01Gate(UnitaryGate):
    """
    Class for representing a CPhase01 gate, a variant of CPhase that affects state ``|01>``

    ::

        CPHASE01(theta) = [[1, 0,            0, 0],
                           [0, exp(i*theta), 0, 0],
                           [0, 0,            1, 0],
                           [0, 0,            0, 1]]
    """

    def __init__(self, theta: float):
        """
        Args:
            theta: Phase angle
        """
        super().__init__(CPHASE01(theta), "cphase01")


class CPhase10Gate(UnitaryGate):
    """
    Class for representing a CPhase10 gate, a variant of CPhase that affects state ``|10>``

    ::

        CPHASE01(theta) = [[1, 0, 0,            0],
                           [0, 1, 0,            0],
                           [0, 0, exp(i*theta), 0],
                           [0, 0, 0,            1]]
    """

    def __init__(self, theta: float):
        """
        Args:
            theta: Phase angle
        """
        super().__init__(CPHASE10(theta), "cphase10")



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Any

from qiskit import QuantumCircuit
from qiskit.circuit import InstructionSet

from .gates import (
    CanonicalGate,
    CPhase00Gate,
    CPhase01Gate,
    CPhase10Gate,
    PSwapGate,
    XYGate,
)


class QuilCircuit(QuantumCircuit):
    """
    A :class:`qiskit.circuit.QuantumCircuit` extension with added support for standard Quil gates:

    https://github.com/rigetti/quilc/blob/master/src/quil/stdgates.quil
    """

    def xy(self, theta: float, qubit1: Any, qubit2: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.xy.XYGate`."""
        return self.append(XYGate(theta), [qubit1, qubit2], [])

    def piswap(self, theta: float, qubit1: Any, qubit2: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.xy.XYGate`."""
        return self.xy(theta, qubit1, qubit2)

    def pswap(self, theta: float, qubit1: Any, qubit2: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.pswap.PSwapGate`."""
        return self.append(PSwapGate(theta), [qubit1, qubit2], [])

    def cphase00(self, theta: float, control_qubit: Any, target_qubit: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.cphase.CPhase00`."""
        return self.append(CPhase00Gate(theta), [control_qubit, target_qubit], [])

    def cphase01(self, theta: float, control_qubit: Any, target_qubit: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.cphase.CPhase01`."""
        return self.append(CPhase01Gate(theta), [control_qubit, target_qubit], [])

    def cphase10(self, theta: float, control_qubit: Any, target_qubit: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.cphase.CPhase10`."""
        return self.append(CPhase10Gate(theta), [control_qubit, target_qubit], [])

    def can(self, alpha: float, beta: float, gamma: float, qubit1: Any, qubit2: Any) -> InstructionSet:
        """Apply :class:`qiskit_rigetti.gates.can.CanonicalGate`."""
        return self.append(CanonicalGate(alpha, beta, gamma), [qubit1, qubit2], [])



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import sys

from ._quil_circuit import QuilCircuit
from ._qcs_backend import RigettiQCSBackend
from ._qcs_job import RigettiQCSJob
from ._qcs_provider import RigettiQCSProvider

if sys.version_info < (3, 8):
    from importlib_metadata import version  # pragma: nocover
else:
    from importlib.metadata import version  # pragma: nocover

__version__ = version(__package__)  # type: ignore



from typing import Callable

from pyquil import Program
from pyquil.gates import RESET

PreExecutionHook = Callable[[Program], Program]
"""Represents a function that can transform a Quil program just before execution."""


def enable_active_reset(quil: Program) -> Program:
    """
    Enable active reset for all qubits.

    See: https://github.com/quil-lang/quil/blob/master/spec/Quil.md#state-reset for more information.

    Args:
        quil: Quil program to transform.

    Returns:
        Copy of the input program, with active reset enabled.
    """
    return quil.prepend_instructions([RESET()])



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings
from collections import Counter
from datetime import datetime
from typing import Optional, Dict, Any, List, Union, Iterator, cast

import numpy as np
from dateutil.tz import tzutc
from pyquil import Program
from pyquil.api import QuantumComputer
from pyquil.api._qpu import QPUExecuteResponse
from pyquil.api._qvm import QVMExecuteResponse
from pyquil.quilbase import RawInstr
from qiskit import QuantumCircuit
from qiskit.providers import JobStatus, JobV1, Backend
from qiskit.providers.models import QasmBackendConfiguration
from qiskit.qobj import QobjExperimentHeader
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from .hooks.pre_compilation import PreCompilationHook
from .hooks.pre_execution import PreExecutionHook

Response = Union[QVMExecuteResponse, QPUExecuteResponse]


class RigettiQCSJob(JobV1):
    """
    Class for representing execution jobs sent to Rigetti backends.
    """

    def __init__(
        self,
        *,
        job_id: str,
        circuits: List[QuantumCircuit],
        options: Dict[str, Any],
        qc: QuantumComputer,
        backend: Backend,
        configuration: QasmBackendConfiguration,
    ) -> None:
        """
        Args:
            job_id: Unique identifier for this job
            circuits: List of circuits to execute
            options: Execution options (e.g. "shots")
            qc: Quantum computer to run against
            backend: :class:`RigettiQCSBackend` that created this job
            configuration: Configuration from parent backend
        """
        super().__init__(backend, job_id)

        self._status = JobStatus.INITIALIZING
        self._circuits = circuits
        self._options = options
        self._qc = qc
        self._configuration = configuration
        self._result: Optional[Result] = None
        self._responses: List[Response] = []

        self._start()

    def submit(self) -> None:
        """
        Raises:
            NotImplementedError: This class uses the asynchronous pattern, so this method should not be called.
        """
        raise NotImplementedError("'submit' is not implemented as this class uses the asynchronous pattern")

    def _start(self) -> None:
        self._responses = [self._start_circuit(circuit) for circuit in self._circuits]
        self._status = JobStatus.RUNNING

    def _start_circuit(self, circuit: QuantumCircuit) -> Response:
        shots = self._options["shots"]
        qasm = circuit.qasm()
        qasm = self._handle_barriers(qasm, circuit.num_qubits)

        before_compile: List[PreCompilationHook] = self._options.get("before_compile", [])
        for fn in before_compile:
            qasm = fn(qasm)

        program = Program(RawInstr(qasm)).wrap_in_numshots_loop(shots)
        program = self._qc.compiler.quil_to_native_quil(program)

        before_execute: List[PreExecutionHook] = self._options.get("before_execute", [])
        for fn in before_execute:
            program = fn(program)

        if self._options.get("ensure_native_quil") and len(before_execute) > 0:
            program = self._qc.compiler.quil_to_native_quil(program)

        executable = self._qc.compiler.native_quil_to_executable(program)

        # typing: QuantumComputer's inner QAM is generic, so we set the expected type here
        return cast(Response, self._qc.qam.execute(executable))

    @staticmethod
    def _handle_barriers(qasm: str, num_circuit_qubits: int) -> str:
        lines = []
        for line in qasm.splitlines():
            if line.startswith("barrier"):
                warnings.warn("barriers are currently omitted during execution on a RigettiQCSBackend")
                continue

            lines.append(line)
        return "\n".join(lines)

    def result(self) -> Result:
        """
        Wait until the job is complete, then return a result.

        Raises:
            JobError: If there was a problem running the Job or retrieving the result
        """
        if self._result is not None:
            return self._result

        now = datetime.now(tzutc())

        results = []
        success = True
        for r in self._get_experiment_results():
            results.append(r)
            success &= r.success

        self._result = Result(
            backend_name=self._configuration.backend_name,
            backend_version=self._configuration.backend_version,
            qobj_id="",
            job_id=self.job_id(),
            success=success,
            results=results,
            date=now,
        )

        self._status = JobStatus.DONE if success else JobStatus.ERROR
        return self._result

    def _get_experiment_results(self) -> Iterator[ExperimentResult]:
        shots = self._options["shots"]

        for circuit_idx, response in enumerate(self._responses):
            states = self._qc.qam.get_result(response).readout_data["ro"]
            memory = list(map(_to_binary_str, np.array(states)))
            success = True
            status = "Completed successfully"

            circuit = self._circuits[circuit_idx]
            yield ExperimentResult(
                header=QobjExperimentHeader(name=circuit.name),
                shots=shots,
                success=success,
                status=status,
                data=ExperimentResultData(counts=Counter(memory), memory=memory),
            )

    def cancel(self) -> None:
        """
        Raises:
            NotImplementedError: There is currently no way to cancel this job.
        """
        raise NotImplementedError("Cancelling jobs is not supported")

    def status(self) -> JobStatus:
        """Get the current status of this Job"""

        return self._status


def _to_binary_str(state: List[int]) -> str:
    # NOTE: According to https://arxiv.org/pdf/1809.03452.pdf, this should be a hex string
    # but it results in missing leading zeros in the displayed output, and binary strings
    # seem to work too. Hex string could be accomplished with:
    #     hex(int(binary_str, 2))
    binary_str = "".join(map(str, state[::-1]))
    return binary_str



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
__all__ = ["XYGate"]

from pyquil.simulation.matrices import XY
from qiskit.extensions import UnitaryGate


class XYGate(UnitaryGate):
    """
    Class for representing an XY gate (parametric iSwap gate)

    ::

        XY(theta) = [[1, 0,                0,                0],
                     [0, cos(theta/2),     i * sin(theta/2), 0],
                     [0, i * sin(theta/2), cos(theta/2),     0],
                     [0, 0,                0,                1]]

    See https://arxiv.org/pdf/1912.04424.pdf for technical details.
    """

    def __init__(self, theta: float):
        """
        Args:
            theta: Phase angle
        """
        super().__init__(XY(theta), "xy")



##############################################################################
# Copyright 2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Any, Optional, List, Dict, cast

import httpx
from pyquil.api import EngagementManager
from qcs_api_client.client import build_sync_client, QCSClientConfiguration
from qiskit.providers import ProviderV1
from qiskit.providers.models import QasmBackendConfiguration

from ._qcs_backend import RigettiQCSBackend


class RigettiQCSProvider(ProviderV1):
    """
    Class for representing the set of Rigetti backends.
    """

    def __init__(
        self,
        *,
        compiler_timeout: float = 10.0,
        execution_timeout: float = 10.0,
        client_configuration: Optional[QCSClientConfiguration] = None,
        engagement_manager: Optional[EngagementManager] = None,
    ) -> None:
        """
        Args:
            execution_timeout: Time limit for execution requests, in seconds.
            compiler_timeout: Time limit for compiler requests, in seconds.
            client_configuration: QCS client configuration. If one is not provided, a default will be loaded.
            engagement_manager: QPU engagement manager. If one is not provided, a default one will be created.
        """
        super().__init__()
        self._backends: List[RigettiQCSBackend] = []
        self._compiler_timeout = compiler_timeout
        self._execution_timeout = execution_timeout
        self._client_configuration = client_configuration or QCSClientConfiguration.load()
        self._engagement_manager = engagement_manager or EngagementManager(
            client_configuration=self._client_configuration
        )

    def backends(self, name: Optional[str] = None, **__: Any) -> List[RigettiQCSBackend]:
        """
        Get the list of :class:`RigettiQCSBackend` corresponding to the available Quantum Processors.

        Args:
            name: An optional QPU name to match against (e.g. "Aspen-9"). If provided, only matching backends will be
                    returned.

        Returns:
            List[RigettiQCSBackend]: The list of matching backends.
        """
        if not self._backends:
            for qpu, data in self._get_quantum_processors().items():
                configuration = _configuration(qpu, num_qubits=data["num_qubits"], local=False, simulator=False)
                self._backends.append(
                    RigettiQCSBackend(
                        compiler_timeout=self._compiler_timeout,
                        execution_timeout=self._execution_timeout,
                        client_configuration=self._client_configuration,
                        engagement_manager=self._engagement_manager,
                        backend_configuration=configuration,
                        provider=self,
                    )
                )

        if name is None:
            return self._backends
        return [b for b in self._backends if b.name() == name]

    def get_simulator(self, *, num_qubits: int, noisy: bool = False) -> RigettiQCSBackend:
        """
        Get a simulator (QVM).

        Args:
            num_qubits: Number of qubits the simulator should have
            noisy: Whether or not the simulator should simulate noise
        Returns:
            RigettiQCSBackend: A backend representing the simulator
        """
        qvm_url = self._client_configuration.profile.applications.pyquil.qvm_url
        local = qvm_url == "" or qvm_url.startswith("http://localhost") or qvm_url.startswith("http://127.0.0.1")
        noisy_str = "-noisy" if noisy else ""
        name = f"{num_qubits}q{noisy_str}-qvm"
        configuration = _configuration(name, num_qubits, local=local, simulator=True)
        return RigettiQCSBackend(
            compiler_timeout=self._compiler_timeout,
            execution_timeout=self._execution_timeout,
            client_configuration=self._client_configuration,
            engagement_manager=self._engagement_manager,
            backend_configuration=configuration,
            provider=self,
        )

    def _get_quantum_processors(self) -> Dict[str, Any]:
        with build_sync_client(configuration=self._client_configuration) as qcs_client:  # type: httpx.Client
            return cast(
                Dict[str, Any], qcs_client.get("https://forest-server.qcs.rigetti.com/devices").json()["devices"]
            )


def _configuration(name: str, num_qubits: int, local: bool, simulator: bool) -> QasmBackendConfiguration:
    return QasmBackendConfiguration(
        backend_name=name,
        backend_version="",
        n_qubits=num_qubits,
        basis_gates=["u1", "u2", "u3", "cx", "id"],
        gates=[],
        local=local,
        simulator=simulator,
        conditional=False,
        open_pulse=False,
        memory=False,
        max_shots=10000,
        coupling_map=[],
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qiskit_rigetti', 'qiskit_rigetti.gates', 'qiskit_rigetti.hooks']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.1,<2.0.0', 'pyquil>=3.0.0,<4.0.0', 'qiskit>=0.27.0,<0.28.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata'],
 'docs': ['sphinx>=4.1.1,<5.0.0',
          'sphinx-autoapi>=1.8.1,<2.0.0',
          'furo>=2021.7.5-beta.38,<2022.0.0',
          'myst-parser>=0.15.1,<0.16.0',
          'sphinx-autobuild>=2021.3.14,<2022.0.0',
          'nbsphinx>=0.8.6,<0.9.0',
          'ipython>=7.25.0,<8.0.0']}

setup_kwargs = {
    'name': 'qiskit-rigetti',
    'version': '0.4.2',
    'description': 'Provider for running Qiskit circuits on Rigetti QPUs and simulators.',
    'long_description': '[![Tests](https://github.com/rigetti/qiskit-rigetti/actions/workflows/test.yml/badge.svg)](https://github.com/rigetti/qiskit-rigetti/actions/workflows/test.yml)\n[![Documentation Status](https://readthedocs.org/projects/qiskit-rigetti/badge/?version=latest)](https://qiskit-rigetti.readthedocs.io/en/latest/?badge=latest)\n[![pypi](https://img.shields.io/pypi/v/qiskit-rigetti.svg)](https://pypi.org/project/qiskit-rigetti/)\n\n# Rigetti Provider for Qiskit\n\n## Pre-requisites\n\n1. Install [Docker](https://www.docker.com/products/docker-desktop)\n1. Download [qelib1.inc](https://raw.githubusercontent.com/Qiskit/qiskit-terra/0.16.2/qiskit/qasm/libs/qelib1.inc)\n1. Place `qelib1.inc` in a folder called `inc` in the project root\n\n## Setup QVM and quilc\n\n### Using Docker Compose\n\nRun `docker compose up` to see service logs or `docker compose up -d` to run in the background.\n\n### Using Docker Manually\n\n1. Start the QVM:\n   \n   ```bash\n   docker run --rm -it -p 5000:5000 rigetti/qvm -S\n   ```\n\n1. Start the compiler:\n\n   ```bash\n   docker run --rm -it -p 5555:5555 -v "$PWD"/inc:/inc rigetti/quilc -S -P --safe-include-directory /inc/\n   ```\n\n## Usage\n\nExample:\n\n```python\nfrom qiskit import execute\nfrom qiskit_rigetti import RigettiQCSProvider, QuilCircuit\n\n# Get provider and backend\np = RigettiQCSProvider()\nbackend = p.get_simulator(num_qubits=2, noisy=True)  # or p.get_backend(name=\'Aspen-9\')\n\n# Create a Bell state circuit\ncircuit = QuilCircuit(2, 2)\ncircuit.h(0)\ncircuit.cx(0, 1)\ncircuit.measure([0, 1], [0, 1])\n\n# Execute the circuit on the backend\njob = execute(circuit, backend, shots=10)\n\n# Grab results from the job\nresult = job.result()\n\n# Return memory and counts\nmemory = result.get_memory(circuit)\ncounts = result.get_counts(circuit)\nprint("Result memory:", memory)\nprint("Result counts:", counts)\n```\n\n### Rigetti Quantum Cloud Services (QCS)\n\nExecution against a QPU requires a [reservation via QCS](https://docs.rigetti.com/qcs/guides/reserving-time-on-a-qpu).\nFor more information on using QCS, see the [QCS documentation](https://docs.rigetti.com).\n\n## Advanced\n\n### Lifecycle Hooks\n\nFor advanced QASM and Quil manipulation, `before_compile` and `before_execute` keyword arguments can be passed to\n`RigettiQCSBackend.run()` or to Qiskit\'s `execute()`.\n\n#### Pre-compilation Hooks\n\nAny `before_compile` hooks will apply, in order, just before compilation from QASM to native Quil.\nFor example:\n\n```python\n...\n\ndef custom_hook_1(qasm: str) -> str:\n   new_qasm = ...\n   return new_qasm\n\ndef custom_hook_2(qasm: str) -> str:\n   new_qasm = ...\n   return new_qasm\n\njob = execute(circuit, backend, shots=10, before_compile=[custom_hook_1, custom_hook_2])\n\n...\n```\n\n#### Pre-execution Hooks\n\nAny `before_execute` hooks will apply, in order, just before execution (after translation from QASM to native Quil).\nFor example:\n\n```python\nfrom pyquil import Program\n\n...\n\ndef custom_hook_1(quil: Program) -> Program:\n   new_quil = ...\n   return new_quil\n\ndef custom_hook_2(quil: Program) -> Program:\n   new_quil = ...\n   return new_quil\n\njob = execute(circuit, backend, shots=10, before_execute=[custom_hook_1, custom_hook_2])\n\n...\n```\n\n> **Note**:\n> \n> Only [certain forms of Quil can can be executed on a QPU](https://pyquil-docs.rigetti.com/en/stable/compiler.html?highlight=protoquil#legal-compiler-input).\n> If pre-execution transformations produce a final program that is not QPU-compliant, `ensure_native_quil=True` can be\n> passed to `execute()` or `RigettiQCSBackend.run()` to recompile the final Quil program to native Quil prior to\n> execution. If no pre-execution hooks were supplied, this setting is ignored. If this setting is omitted, a value of\n> `False` is assumed.\n> \n> _Example_: Adding the Quil instruction `H 0` would result in an error if `ensure_native_quil=False` and the QPU does\n> not natively implement Hadamard gates.\n\n#### Built-in Hooks\n\nThe `hooks.pre_compilation` and `hooks.pre_execution` packages provide a number of convenient hooks:\n\n##### `set_rewiring`\n\nUse `set_rewiring` to provide a [rewiring directive](https://pyquil-docs.rigetti.com/en/stable/compiler.html#initial-rewiring)\nto the Quil compiler. For example:\n\n```python\nfrom qiskit_rigetti.hooks.pre_compilation import set_rewiring\n\n...\n\njob = execute(circuit, backend, shots=10, before_compile=[set_rewiring("NAIVE")])\n\n...\n```\n\n> **Note**: Rewiring directives require `quilc` version 1.25 or higher.\n\n##### `enable_active_reset`\n\nUse `enable_active_reset` to enable [active qubit reset](https://github.com/quil-lang/quil/blob/master/spec/Quil.md#state-reset),\nan optimization that can significantly reduce the time between executions. For example:\n\n```python\nfrom qiskit_rigetti.hooks.pre_execution import enable_active_reset\n\n...\n\njob = execute(circuit, backend, shots=10, before_execute=[enable_active_reset])\n\n...\n```\n\n## Development\n\n> **Note**: This module is developed in Python 3.7, other versions will currently fail type checking.\n\nDependencies are managed with [Poetry](https://python-poetry.org/) so you need to install that first. Once you\'ve installed all dependencies (`poetry install`) and activated the virtual environment (`poetry shell`), you can use these rules from the `Makefile` to run common tasks:\n\n1. Run tests: `make test`\n1. Check style and types: `make check-all`\n1. Check style only: `make check-style`\n1. Check types only: `make check-types`\n1. Reformat all code (to make `check-style` pass): `make format`\n1. Build documentation, serve locally, and watch for changes: `make watch-docs` (requires `docs` extra: `poetry install -E docs`)\n',
    'author': 'Rigetti Computing',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rigetti/qiskit-rigetti',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
